"""
These views are intended for use in rendering email templates
within the admin site, and supporting preview functionality.
"""
import json
import logging

from django.conf import settings as django_settings
from django.contrib.auth.decorators import user_passes_test
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.views.decorators.clickjacking import xframe_options_sameorigin
from django.urls import reverse

from .forms import MultiEmailTemplateField, EmailTestForm
from .helpers import merge_dicts
from .models import EmailTemplate

logger = logging.getLogger(__name__)


@user_passes_test(lambda u: u.is_staff)
@xframe_options_sameorigin
def render_template_subject(request, template_id):
    """Render the template subject."""
    template = get_object_or_404(EmailTemplate, id=template_id)
    html = template.render_subject(template.test_context)
    return HttpResponse(html, content_type='text/plain')


@user_passes_test(lambda u: u.is_staff)
@xframe_options_sameorigin
def render_template_body(request, template_id, content_type):
    """Render the template body as plain text or HTML."""
    template = get_object_or_404(EmailTemplate, id=template_id)
    if content_type in (EmailTemplate.CONTENT_TYPE_PLAIN, EmailTemplate.CONTENT_TYPE_HTML):
        html = template.render_body(template.test_context, content_type)
        return HttpResponse(html, content_type=content_type)
    # do not return the content_type to the user, as it is
    # user-generated and _could_ be a vulnerability.
    return HttpResponse("Invalid content_type specified.", status=400)


@user_passes_test(lambda u: u.is_staff)
def send_test_email(request):
    """Intermediate admin action page for sending a single test email."""
    # use the field.to_python here as belt-and-braces - if it works here
    # we can be confident that it'll work on the POST.
    templates = MultiEmailTemplateField().to_python(request.GET['templates'])

    if request.method == 'GET':
        contexts = merge_dicts(*[t.test_context for t in templates])
        context = json.dumps(contexts, indent=4, sort_keys=True)
        initial = {
            'templates': request.GET['templates'],
            'context': context
        }
        try:
            template = templates.get()
            initial['from_email'] = template.from_email
            initial['reply_to'] = template.reply_to
        except EmailTemplate.MultipleObjectsReturned:
            initial['from_email'] = django_settings.DEFAULT_FROM_EMAIL
            initial['reply_to'] = django_settings.DEFAULT_FROM_EMAIL
        form = EmailTestForm(initial=initial)
        return render(
            request,
            'appmail/send_test_email.html',
            {
                'form': form,
                'templates': templates,
                # opts are used for rendering some page furniture - breadcrumbs etc.
                'opts': EmailTemplate._meta,
            }
        )

    if request.method == 'POST':
        form = EmailTestForm(request.POST)
        if form.is_valid():
            form.send_emails(request)
            return HttpResponseRedirect(
                '{}?{}'.format(
                    reverse('appmail:send_test_email'),
                    request.GET.urlencode()
                )
            )
        else:
            return render(
                request,
                'appmail/send_test_email.html',
                {
                    'form': form,
                    'templates': templates,
                    # opts are used for rendering some page furniture - breadcrumbs etc.
                    'opts': EmailTemplate._meta,
                },
                status=422
            )



try:
    from django.urls import re_path
except ImportError:
    from django.conf.urls import url as re_path

from .views import (
    render_template_body,
    render_template_subject,
    send_test_email
)

app_name = 'appmail'

urlpatterns = [
    re_path(
        r'^templates/(?P<template_id>\d+)/body.txt$',
        render_template_body,
        kwargs={'content_type': 'text/plain'},
        name="render_template_body_text"
    ),
    re_path(
        r'^templates/(?P<template_id>\d+)/body.html$',
        render_template_body,
        kwargs={'content_type': 'text/html'},
        name="render_template_body_html"
    ),
    re_path(
        r'^templates/(?P<template_id>\d+)/subject.txt$',
        render_template_subject,
        name="render_template_subject"
    ),
    re_path(
        r'^templates/test/$',
        send_test_email,
        name="send_test_email"
    ),
]



from django.conf import settings
from django.contrib.postgres.fields import JSONField
from django.core.exceptions import ValidationError
from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.template import (
    Context,
    Template,
    TemplateDoesNotExist,
    TemplateSyntaxError
)
from django.utils.translation import ugettext_lazy as _lazy

from . import helpers
from .settings import (
    ADD_EXTRA_HEADERS,
    VALIDATE_ON_SAVE,
    CONTEXT_PROCESSORS
)


class EmailTemplateQuerySet(models.query.QuerySet):

    def active(self):
        """Returns active templates only."""
        return self.filter(is_active=True)

    def current(self, name, language=settings.LANGUAGE_CODE):
        """Returns the latest version of a template."""
        return self.active().filter(name=name, language=language).order_by('version').last()

    def version(self, name, version, language=settings.LANGUAGE_CODE):
        """Returns a specific version of a template."""
        return self.active().get(name=name, language=language, version=version)


class EmailTemplate(models.Model):

    """
    Email template. Contains HTML and plain text variants.

    Each Template object has a unique name:language.version combination, which
    means that localisation of templates is managed through having multiple
    objects with the same name - there is no inheritence model. This is to
    keep it simple:

        order-confirmation:en.0
        order-confirmation:de.0
        order-confirmation:fr.0

    Templates contain HTML and plain text content.

    """
    CONTENT_TYPE_PLAIN = 'text/plain'
    CONTENT_TYPE_HTML = 'text/html'
    CONTENT_TYPES = (CONTENT_TYPE_PLAIN, CONTENT_TYPE_HTML)

    name = models.CharField(
        _lazy('Template name'),
        max_length=100,
        help_text=_lazy("Template name - must be unique for a given language/version combination."),
        db_index=True
    )
    description = models.CharField(
        _lazy('Description'),
        max_length=100,
        help_text=_lazy("Optional description. e.g. used to differentiate variants ('new header')."),  # noqa
        blank=True
    )
    # language is free text and not a choices field as we make no assumption
    # as to how the end user is storing / managing languages.
    language = models.CharField(
        _lazy('Language'),
        max_length=20,
        default=settings.LANGUAGE_CODE,
        help_text=_lazy(
            "Used to support localisation of emails, defaults to `settings.LANGUAGE_CODE`, "
            "but can be any string, e.g. 'London', 'NYC'."
        ),
        db_index=True
    )
    version = models.IntegerField(
        _lazy('Version (or variant)'),
        default=0,
        help_text=_lazy("Integer value - can be used for versioning or A/B testing."),
        db_index=True
    )
    subject = models.CharField(
        _lazy('Subject line template'),
        max_length=100,
        help_text=_lazy("Email subject line (may contain template variables)."),
    )
    body_text = models.TextField(
        _lazy('Plain text template'),
        help_text=_lazy("Plain text content (may contain template variables)."),
    )
    body_html = models.TextField(
        _lazy('HTML template'),
        help_text=_lazy("HTML content (may contain template variables)."),
    )
    test_context = JSONField(
        default=dict,
        blank=True,
        help_text=_lazy("Dummy JSON used for test rendering (set automatically on first save).")
    )
    is_active = models.BooleanField(
        _lazy("Active (live)"),
        help_text=_lazy("Set to False to remove from `current` queryset."),
        default=True
    )
    from_email = models.CharField(
        _lazy("Sender"),
        max_length=254,
        help_text=_lazy(
            "Default sender address if none specified. Verbose form is accepted."
        ),
        default=settings.DEFAULT_FROM_EMAIL
    )
    reply_to = models.CharField(
        _lazy("Reply-To"),
        max_length=254,
        help_text=_lazy("Comma separated list of Reply-To recipients."),
        default=settings.DEFAULT_FROM_EMAIL
    )

    objects = EmailTemplateQuerySet().as_manager()

    class Meta:
        unique_together = ("name", "language", "version")

    def __str__(self):
        return "'{}' ({})".format(self.name, self.language)

    def __repr__(self):
        return (
            "<EmailTemplate id={} name='{}' language='{}' version={}>".format(
                self.id, self.name, self.language, self.version
            )
        )

    @property
    def extra_headers(self):
        return{
            'X-Appmail-Template': (
                'name=%s; language=%s; version=%s' % (self.name, self.language, self.version)
            )
        }

    @property
    def reply_to_list(self):
        """Convert the reply_to field to a list."""
        return [a.strip() for a in self.reply_to.split(',')]

    def save(self, *args, **kwargs):
        """Update dummy context on first save and validate template contents.

        Kwargs:
            validate: set to False to bypass template validation; defaults
                to settings.VALIDATE_ON_SAVE.

        """
        if self.pk is None:
            self.test_context = helpers.get_context(
                self.subject +
                self.body_text +
                self.body_html
            )
        validate = kwargs.pop('validate', VALIDATE_ON_SAVE)
        if validate:
            self.clean()
        super(EmailTemplate, self).save(*args, **kwargs)
        return self

    def clean(self):
        """Validate model - specifically that the template can be rendered."""
        validation_errors = {}
        validation_errors.update(self._validate_body(EmailTemplate.CONTENT_TYPE_PLAIN))
        validation_errors.update(self._validate_body(EmailTemplate.CONTENT_TYPE_HTML))
        validation_errors.update(self._validate_subject())
        if validation_errors:
            raise ValidationError(validation_errors)

    def render_subject(self, context, processors=CONTEXT_PROCESSORS):
        """Render subject line."""
        ctx = Context(helpers.patch_context(context, processors), autoescape=False)
        return Template(self.subject).render(ctx)

    def _validate_subject(self):
        """Try rendering the body template and capture any errors."""
        try:
            self.render_subject({})
        except TemplateDoesNotExist as ex:
            return {'subject': _lazy("Template does not exist: {}".format(ex))}
        except TemplateSyntaxError as ex:
            return {'subject': str(ex)}
        else:
            return {}

    def render_body(self, context, content_type=CONTENT_TYPE_PLAIN, processors=CONTEXT_PROCESSORS):
        """Render email body in plain text or HTML format."""
        assert content_type in EmailTemplate.CONTENT_TYPES, _lazy("Invalid content type.")
        if content_type == EmailTemplate.CONTENT_TYPE_PLAIN:
            ctx = Context(helpers.patch_context(context, processors), autoescape=False)
            return Template(self.body_text).render(ctx)
        if content_type == EmailTemplate.CONTENT_TYPE_HTML:
            ctx = Context(helpers.patch_context(context, processors))
            return Template(self.body_html).render(ctx)

    def _validate_body(self, content_type):
        """Try rendering the body template and capture any errors."""
        assert content_type in EmailTemplate.CONTENT_TYPES, _lazy("Invalid content type.")
        if content_type == EmailTemplate.CONTENT_TYPE_PLAIN:
            field_name = 'body_text'
        if content_type == EmailTemplate.CONTENT_TYPE_HTML:
            field_name = 'body_html'
        try:
            self.render_body({}, content_type=content_type)
        except TemplateDoesNotExist as ex:
            return {field_name: _lazy("Template does not exist: {}".format(ex))}
        except TemplateSyntaxError as ex:
            return {field_name: str(ex)}
        else:
            return {}

    def create_message(self, context, **email_kwargs):
        """
        Return populated EmailMultiAlternatives object.

        This function is a helper that will render the template subject and
        plain text / html content, as well as populating all of the standard
        EmailMultiAlternatives properties.

            >>> template = EmailTemplate.objects.get_latest('order_summary')
            >>> context = {'first_name': "Bruce", 'last_name'="Lee"}
            >>> email = template.create_message(context, to=['bruce@kung.fu'])
            >>> email.send()

        The function supports all of the standard EmailMultiAlternatives
        constructor kwargs except for 'subject', 'body' and 'alternatives' - as
        these are set from the template (subject, body_text and body_html).

        """
        for kw in ('subject', 'body', 'alternatives'):
            assert kw not in email_kwargs, _lazy("Invalid create_message kwarg: '{}'".format(kw))
        subject = self.render_subject(context)
        body = self.render_body(context, content_type=EmailTemplate.CONTENT_TYPE_PLAIN)
        html = self.render_body(context, content_type=EmailTemplate.CONTENT_TYPE_HTML)
        email_kwargs['reply_to'] = email_kwargs.get('reply_to') or self.reply_to_list
        email_kwargs['from_email'] = email_kwargs.get('from_email') or self.from_email
        if ADD_EXTRA_HEADERS:
            email_kwargs['headers'] = email_kwargs.get('headers', {})
            email_kwargs['headers'].update(self.extra_headers)
        # alternatives is a list of (content, mimetype) tuples
        # https://github.com/django/django/blob/master/django/core/mail/message.py#L435
        return EmailMultiAlternatives(
            subject=subject,
            body=body,
            alternatives=[(html, EmailTemplate.CONTENT_TYPE_HTML)],
            **email_kwargs
        )

    def clone(self):
        """Create a copy of the current object, increase version by 1."""
        self.pk = None
        self.version += 1
        return self.save()




from django.test import TestCase

from .. import helpers


class HelperTests(TestCase):

    """appmail.helpers module tests."""

    def test_get_context(self):
        self.assertEqual(
            helpers.get_context('{{a}} {{b.c}}'),
            {
                'a': "A",
                'b': {
                    'c': "C"
                }
            }
        )

    def test_extract_vars(self):
        """Check extract_vars handles expected input."""
        for x, y in (
            ('', []),
            (None, []),
            ('{{foo}}', ['foo']),
            ('{{ foo}}', ['foo']),
            ('{{ foo }}', ['foo']),
            ('{{ foo }', []),
            ('{% foo %}', []),
            ('{{ foo|time }}', []),
            ('{{foo}} {{bar}}', ['foo', 'bar']),
        ):
            self.assertEqual(set(helpers.extract_vars(x)), set(y))

    def test_expand_list(self):
        """Check dot notation expansion."""
        self.assertRaises(AssertionError, helpers.expand_list, None)
        self.assertRaises(AssertionError, helpers.expand_list, '')
        self.assertEqual(
            helpers.expand_list(['a', 'b.c']),
            {
                'a': {},
                'b': {
                    'c': {}
                }
            }
        )

    def test_fill_leaf_values(self):
        """Check the default func is applied."""
        self.assertRaises(AssertionError, helpers.fill_leaf_values, None)
        self.assertRaises(AssertionError, helpers.fill_leaf_values, '')
        self.assertEqual(
            helpers.fill_leaf_values(
                {
                    'a': {},
                    'b': {
                        'c': {}
                    }
                }
            ),
            {
                'a': "A",
                'b': {
                    'c': "C"
                }
            }
        )

    def test_merge_dicts(self):
        self.assertEqual(
            helpers.merge_dicts({'foo': 1}),
            {'foo': 1}
        )
        self.assertEqual(
            helpers.merge_dicts({'foo': 1}, {'bar': 2}),
            {'foo': 1, 'bar': 2}
        )
        self.assertEqual(
            helpers.merge_dicts({'foo': 1}, {'foo': 2}),
            {'foo': 2}
        )

    def test_patch_context(self):

        foo = {'foo': 1}
        bar = {'bar': 2}
        baz = {'baz': 3}

        def cp1(request):
            return bar

        def cp2(request):
            return baz

        self.assertEqual(
            helpers.patch_context(foo, []),
            foo
        )

        self.assertEqual(
            helpers.patch_context(foo, [cp1]),
            helpers.merge_dicts(foo, bar)
        )

        self.assertEqual(
            helpers.patch_context(foo, [cp1, cp2]),
            helpers.merge_dicts(foo, bar, baz)
        )



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-08-07 06:10
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('appmail', '0003_emailtemplate_test_context'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailtemplate',
            name='is_active',
            field=models.BooleanField(default=True, help_text='Set to False to remove from `current` queryset.', verbose_name='Active (live)'),
        ),
    ]



from unittest import mock

from django.contrib.auth.models import AnonymousUser, User
from django.test import TestCase, RequestFactory
from django.urls import reverse

from .. import views
from ..forms import EmailTestForm
from ..models import EmailTemplate


class ViewTests(TestCase):

    """appmail.helpers module tests."""

    def setUp(self):
        # Every test needs access to the request factory.
        self.factory = RequestFactory()
        self.template = EmailTemplate(
            subject='ßello, {{user.first_name}}',
            body_text='ßello,\n{{user.first_name}}',
            body_html="ßello, <b>{{user.first_name}}</b>"
        ).save()
        self.subject_url = reverse(
            'appmail:render_template_subject',
            kwargs={'template_id': self.template.id}
        )
        self.body_text_url = reverse(
            'appmail:render_template_body_text',
            kwargs={'template_id': self.template.id}
        )
        self.body_html_url = reverse(
            'appmail:render_template_body_html',
            kwargs={'template_id': self.template.id}
        )

    def test_render_template_subject(self):
        template = self.template
        request = self.factory.get(self.subject_url)
        # check that non-staff are denied
        request.user = AnonymousUser()
        response = views.render_template_subject(request, template.id)
        self.assertEqual(response.status_code, 302)
        # non-staff user
        request.user = User()
        response = views.render_template_subject(request, template.id)
        self.assertEqual(response.status_code, 302)
        # staff user
        request.user.is_staff = True
        response = views.render_template_subject(request, template.id)
        self.assertEqual(response.status_code, 200)
        # should render the template with the dummy context
        self.assertEqual(
            response.content.decode('utf-8'),
            template.render_subject(template.test_context)
        )

    def test_render_template_body(self):
        template = self.template
        request = self.factory.get(self.body_text_url)
        # check that non-staff are denied
        request.user = AnonymousUser()
        response = views.render_template_body(
            request,
            template.id,
            EmailTemplate.CONTENT_TYPE_PLAIN
        )
        self.assertEqual(response.status_code, 302)
        # non-staff user
        request.user = User()
        response = views.render_template_body(
            request,
            template.id,
            EmailTemplate.CONTENT_TYPE_PLAIN
        )
        self.assertEqual(response.status_code, 302)
        # staff user
        request.user.is_staff = True
        response = views.render_template_body(
            request,
            template.id,
            EmailTemplate.CONTENT_TYPE_PLAIN
        )
        self.assertEqual(response.status_code, 200)
        # should render the template with the dummy context
        self.assertEqual(
            response.content.decode('utf-8'),
            template.render_body(
                template.test_context,
                EmailTemplate.CONTENT_TYPE_PLAIN
            )
        )
        # now check the HTML version
        response = views.render_template_body(
            request,
            template.id,
            EmailTemplate.CONTENT_TYPE_HTML
        )
        self.assertEqual(
            response.content.decode('utf-8'),
            template.render_body(
                template.test_context,
                EmailTemplate.CONTENT_TYPE_HTML
            )
        )

        # now check the HTML version
        response = views.render_template_body(
            request,
            template.id,
            "foo"
        )
        self.assertEqual(response.status_code, 400)
        self.assertEqual(
            response.content.decode('utf-8'),
            "Invalid content_type specified."
        )

    def test_send_test_emails_GET(self):
        user = User.objects.create(username='admin', password='password', is_staff=True)
        template = self.template
        url = '{}?templates={}'.format(
            reverse('appmail:send_test_email'),
            template.pk
        )
        response = self.client.get(url)
        self.assertEqual(response.status_code, 302)
        self.client.force_login(user)
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, template.name)

    @mock.patch.object(EmailTestForm, 'send_emails')
    def test_send_test_emails_POST(self, mock_send):
        user = User.objects.create(username='admin', password='password', is_staff=True)
        template = self.template
        url = '{}?templates={}'.format(
            reverse('appmail:send_test_email'),
            template.pk
        )
        self.client.force_login(user)
        payload = {
            'to': 'fred@example.com',
            'cc': '',
            'bcc': '',
            'context': '',
            'from_email': 'donotreply@example.com',
            'reply_to': ['donotreply1@example.com'],
            'templates': template.pk
        }
        response = self.client.post(url, payload)
        self.assertEqual(mock_send.call_count, 1)
        mock_send.assert_called_once_with(response.wsgi_request)
        self.assertEqual(response.status_code, 302)

        # check that bad response returns 422
        response = self.client.post(url, {})
        self.assertEqual(mock_send.call_count, 1)
        self.assertEqual(response.status_code, 422)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-05-06 08:15
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('appmail', '0002_add_template_description'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailtemplate',
            name='test_context',
            field=django.contrib.postgres.fields.jsonb.JSONField(
                default=dict,
                blank=True,
                help_text='Dummy JSON used for test rendering (set automatically on first save).')
            ,
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2017-09-11 03:58
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('appmail', '0004_emailtemplate_is_active'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailtemplate',
            name='from_email',
            field=models.CharField(default=settings.DEFAULT_FROM_EMAIL, help_text='Default sender address if none specified. Verbose form is accepted.', max_length=254, verbose_name='Sender'),
        ),
        migrations.AddField(
            model_name='emailtemplate',
            name='reply_to',
            field=models.CharField(default=settings.DEFAULT_FROM_EMAIL, help_text='Comma separated list of Reply-To recipients.', max_length=254, verbose_name='Reply-To'),
        ),
        migrations.AlterField(
            model_name='emailtemplate',
            name='description',
            field=models.CharField(blank=True, help_text="Optional description. e.g. used to differentiate variants ('new header').", max_length=100, verbose_name='Description'),
        ),
    ]



from django.contrib import admin
from django.contrib import messages
from django.contrib.postgres.fields import JSONField
from django.core.exceptions import ValidationError
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import ugettext_lazy as _

from .forms import JSONWidget
from .models import EmailTemplate


class ValidTemplateListFilter(admin.SimpleListFilter):

    """Filter on whether the template can be rendered or not."""

    title = _('Is valid')
    parameter_name = 'valid'

    def lookups(self, request, model_admin):
        """
        Returns a list of tuples. The first element in each
        tuple is the coded value for the option that will
        appear in the URL query. The second element is the
        human-readable name for the option that will appear
        in the right sidebar.
        """
        return (
            ('1', _('True')),
            ('0', _('False')),
        )

    def queryset(self, request, queryset):
        """
        Returns the filtered queryset based on the value
        provided in the query string and retrievable via
        `self.value()`.
        """
        valid_ids = []
        invalid_ids = []
        for obj in queryset:
            try:
                obj.clean()
                valid_ids.append(obj.pk)
            except ValidationError:
                invalid_ids.append(obj.pk)

        if self.value() == '1':
            return queryset.filter(pk__in=valid_ids)
        if self.value() == '0':
            return queryset.filter(pk__in=invalid_ids)


class EmailTemplateAdmin(admin.ModelAdmin):

    formfield_overrides = {
        JSONField: {'widget': JSONWidget},
    }

    list_display = (
        'name',
        'subject',
        'language',
        'version',
        'has_text',
        'has_html',
        'is_valid',
        'is_active'
    )

    list_filter = (
        'language',
        'version',
        ValidTemplateListFilter,
        'is_active'
    )

    readonly_fields = (
        'render_subject',
        'render_text',
        'render_html',
    )

    search_fields = (
        'name',
        'subject'
    )
    actions = (
        'activate_templates',
        'deactivate_templates',
        'clone_templates',
        'send_test_emails',
    )
    fieldsets = (
        (
            'Basic Information',
            {
                'fields': (
                    'name',
                    'description',
                    'language',
                    'version',
                    'is_active',
                )
            }
        ),
        (
            'Email Defaults',
            {
                'fields': (
                    'from_email',
                    'reply_to',
                )
            }
        ),
        (
            'Templates',
            {
                'fields': (
                    'subject',
                    'body_text',
                    'body_html',
                )
            }
        ),
        (
            'Sample Output',
            {
                'fields': (
                    'test_context',
                    'render_subject',
                    'render_text',
                    'render_html',
                )
            }
        )
    )

    def _iframe(self, url):
        return format_html(
            "<iframe class='appmail' src='{}' onload='resizeIframe(this)'></iframe><br/>"
            "<a href='{}' target='_blank'>View in new tab.</a>",
            url,
            url
        )

    # these functions are here rather than on the model so that we can get the
    # boolean icon.
    def has_text(self, obj):
        return len(obj.body_text or '') > 0
    has_text.boolean = True

    def has_html(self, obj):
        return len(obj.body_html or '') > 0
    has_html.boolean = True

    def is_valid(self, obj):
        """Return True if the template can be rendered."""
        try:
            obj.clean()
            return True
        except ValidationError:
            return False
    is_valid.boolean = True

    def render_subject(self, obj):
        if obj.id is None:
            url = ''
        else:
            url = reverse(
                'appmail:render_template_subject',
                kwargs={'template_id': obj.id}
            )
        return self._iframe(url)
    render_subject.short_description = 'Rendered subject'
    render_subject.allow_tags = True

    def render_text(self, obj):
        if obj.id is None:
            url = ''
        else:
            url = reverse(
                'appmail:render_template_body_text',
                kwargs={'template_id': obj.id}
            )
        return self._iframe(url)
    render_text.short_description = 'Rendered body (plain)'
    render_text.allow_tags = True

    def render_html(self, obj):
        if obj.id is None:
            url = ''
        else:
            url = reverse(
                'appmail:render_template_body_html',
                kwargs={'template_id': obj.id}
            )
        return self._iframe(url)
    render_html.short_description = 'Rendered body (html)'
    render_html.allow_tags = True

    def send_test_emails(self, request, queryset):
        selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
        url = '{}?templates={}'.format(
            reverse('appmail:send_test_email'),
            ','.join(selected)
        )
        return HttpResponseRedirect(url)
    send_test_emails.short_description = _("Send test email for selected templates")

    def clone_templates(self, request, queryset):
        selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
        templates = EmailTemplate.objects.filter(pk__in=selected)
        for template in templates:
            template.clone()
            messages.success(request, _("Cloned template '%s'" % template.name))
        return HttpResponseRedirect(request.path)
    clone_templates.short_description = _("Clone selected email templates")

    def activate_templates(self, request, queryset):
        selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
        templates = EmailTemplate.objects.filter(pk__in=selected)
        count = templates.update(is_active=True)
        messages.success(request, _("Activated %s templates" % count))
        return HttpResponseRedirect(request.path)
    activate_templates.short_description = _("Activate selected email templates")

    def deactivate_templates(self, request, queryset):
        selected = request.POST.getlist(admin.ACTION_CHECKBOX_NAME)
        templates = EmailTemplate.objects.filter(pk__in=selected)
        count = templates.update(is_active=False)
        messages.success(request, _("Deactivated %s templates" % count))
        return HttpResponseRedirect(request.path)
    deactivate_templates.short_description = _("Deactivate selected email templates")


admin.site.register(EmailTemplate, EmailTemplateAdmin)




from django.conf import settings
from django.utils.module_loading import import_string

# Validate that EmailTemplate can be rendered without
# error on each save. Defaults to True.
VALIDATE_ON_SAVE = getattr(settings, 'APPMAIL_VALIDATE_ON_SAVE', True)
# if True then add X-Appmail-* headers to outgoung email objects
ADD_EXTRA_HEADERS = getattr(settings, 'APPMAIL_ADD_HEADERS', True)
# list of context processor functions applied on each render
CONTEXT_PROCESSORS = [import_string(s) for s in getattr(settings, 'APPMAIL_CONTEXT_PROCESSORS', [])]  # noqa



import re
from django.utils.safestring import mark_safe

# regex for extracting django template {{ variable }}s
TEMPLATE_VARS = re.compile(r'{{([ ._[a-z]*)}}')


def get_context(content):
    """
    Return a dummary context dict for a content block.

    This function works by taking template content,
    extracting the template variables ({{ foo.bar }}),
    expanding out the list of variables into a dict
    using the '.' separator, and then populating the
    value of each leaf node with the node key ('foo': "FOO").

    Used for generating test data.

    """
    return fill_leaf_values(expand_list(extract_vars(content)))


def extract_vars(content):
    """
    Extract variables from template content.

    Returns a deduplicated list of all the variable names
    found in the content.

    """
    content = content or ''
    # if I was better at regex I wouldn't need the strip.
    return list(set([s.strip() for s in TEMPLATE_VARS.findall(content)]))


def expand_list(_list):
    """
    Convert list of '.' separated values to a nested dict.

    Taken from SO article which I now can't find, this will take a list
    and return a dictionary which contains an empty dict as each leaf
    node.

        >>> expand_list(['a', 'b.c'])
        {
            'a': {},
            'b': {
                'c': {}
            }
        }

    """
    assert isinstance(_list, list), "arg must be a list"
    tree = {}
    for item in _list:
        t = tree
        for part in item.split('.'):
            t = t.setdefault(part, {})
    return tree


def fill_leaf_values(tree):
    """
    Recursive function that populates empty dict leaf nodes.

    This function will look for all the leaf nodes in a dictionary
    and replace them with a value that looks like the variable
    in the template - e.g. {{ foo }}.

        >>> fill_leaf_values({'a': {}, 'b': 'c': {}})
        {
            'a': '{{ A }}',
            'b': {
                'c': '{{ C }}'
            }
        }

    """
    assert isinstance(tree, dict), "arg must be a dictionary"
    for k in tree.keys():
        if tree[k] == {}:
            tree[k] = k.upper()
        else:
            fill_leaf_values(tree[k])
    return tree


def merge_dicts(*dicts):
    """Merge multiple dicts into one."""
    context = {}
    for d in dicts:
        context.update(d)
    return context


def patch_context(context, processors, request=None):
    """Add template context_processor content to context."""
    cpx = [p(request) for p in processors]
    return merge_dicts(context, *cpx)




# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2017-03-11 11:36
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='EmailTemplate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, help_text='Template name - must be unique for a given language/version combination.', max_length=100, verbose_name='Template name')),
                ('language', models.CharField(db_index=True, default='en-us', help_text='Used to support localisation of emails, defaults to settings.LANGUAGE_CODE.', max_length=20, verbose_name='Language code')),
                ('version', models.IntegerField(db_index=True, default=0, help_text='Integer value - can be used for versioning or A/B testing.', verbose_name='Version (or variant)')),
                ('subject', models.CharField(help_text='Email subject line (may contain template variables).', max_length=100, verbose_name='Subject line template')),
                ('body_text', models.TextField(help_text='Plain text content (may contain template variables).', verbose_name='Plain text template')),
                ('body_html', models.TextField(help_text='HTML content (may contain template variables).', verbose_name='HTML template')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='emailtemplate',
            unique_together=set([('name', 'language', 'version')]),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.6 on 2017-03-15 03:26
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('appmail', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='emailtemplate',
            name='description',
            field=models.CharField(blank=True, help_text="Optional description. e.g. used to differentiate variants ('new header').", max_length=100),
        ),
        migrations.AlterField(
            model_name='emailtemplate',
            name='language',
            field=models.CharField(db_index=True, default=settings.LANGUAGE_CODE, help_text="Used to support localisation of emails, defaults to `settings.LANGUAGE_CODE`, but can be any string, e.g. 'London', 'NYC'.", max_length=20, verbose_name='Language'),
        ),
    ]
